"""
A device driver implementation for the serial interface of a Sartorius Balance

:author: Florian Meinicke (florian.meinicke@cetoni.de)
:date: 10.01.2022
"""

import logging
import random
import re
import time
from threading import Event, Lock, Thread, Timer
from typing import Optional

from .abc import BalanceInterface, BalanceNotFoundException, SerialBalanceInterface


class SartoriusBalanceNotFoundException(BalanceNotFoundException):
    def __init__(self, port: Optional[str] = None):
        super().__init__(port,
            "No Sartorius balance detected on {port}. Please verify that your balance has the following RS232 "
            "settings: [DAT.REC: SBI], [BAUD: 115200], [PARITY: NONE], [STOPBIT: 1 BIT], [HANDSHK.: NONE], "
            "[DATABITS: 8 BIT]. For best performance, please ensure that you set COM.SBI -> AUTO.CYCL to EACH VAL."
        )


class SartoriusBalance(SerialBalanceInterface):
    """
    Class for reading serial data from a Sartorius balance
    """

    not_found_exception = SartoriusBalanceNotFoundException

    def __init__(self, port: Optional[str] = None):
        super().__init__(port)

    @staticmethod
    def serial_data_to_value(data) -> float:
        # data = 'G     +   0.0006 !  '
        #               ^~~~~~~~~~ match this (and remove spaces later)
        value_regex = re.compile(r"[+-]?\s+\d+\.\d+")

        return float(value_regex.findall(data)[0].replace(" ", ""))

    def tare(self):
        """
        Executes the tare command
        """
        self._protocol.write_line("T")

    @property
    def unique_balance_identifier_request(self) -> str:
        return "x1_"

    @staticmethod
    def is_valid_balance(data) -> bool:
        # b'Model  BCE124I-1CEU \r\n'
        #                      ^~~~~ positive lookahead   (?=\s+\\r\\n.*)?
        #          ^~~~~~~~~~~~ matches this              [\w\d-]+
        #   ^~~~~~~ positive lookbehind                   (?<=Model\s{2})
        model_regex = re.compile(r"(?<=Model\s{2})[\w\d-]+(?=\s+\r\n.*)?")

        return len(model_regex.findall(data)) > 0


class SartoriusBalanceSim(BalanceInterface):
    """
    A simple simulated device driver implementation for Sartorius balances
    """

    __stop_event: Event
    __tare_inhibit_lock: Lock
    __tare_inhibit_timer: Timer

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.__stop_event = Event()
        self.__tare_inhibit_lock = Lock()
        self.__tare_inhibit_timer = Timer(3, self.__tare_inhibit_lock.release)

    def tare(self):
        """
        Executes the tare command
        """
        self._value = 0
        self.__tare_inhibit_timer.start()
        self.__tare_inhibit_lock.acquire()

    def start(self):
        def generate_values():
            while not self.__stop_event.wait(0.5):
                if not self.__tare_inhibit_lock.locked():
                    self._value = random.uniform(-50, 150)

        Thread(target=generate_values).start()

    def stop(self):
        self.__stop_event.set()


# ----------------------------------------------------------------------------
# test
if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(threadName)-12.12s] %(levelname)-8s| %(name)s %(module)s.%(funcName)s: %(message)s",
        level=logging.DEBUG,
    )

    balance = SartoriusBalance()
    balance.open()
    time.sleep(0.3)
    balance.tare()
    time.sleep(5)
