"""
A device driver implementation for the REST API of a Memmert Oven

:author: Florian Meinicke (florian.meinicke@cetoni.de)
:date: 12.09.2022
"""

from __future__ import annotations

import datetime
import json
import logging
import math
import threading
import time
from enum import Enum, IntEnum
from threading import Event, Thread
from typing import Any, Dict, List, NamedTuple, Optional

import requests
from requests import ConnectionError, ConnectTimeout

from .abc import (
    EventLogEntry,
    EventLogEntryType,
    MemmertOvenInterface,
    Range,
    TemperatureControllerInterface,
    TemperatureControllerNotFoundException,
)

logger = logging.getLogger(__name__)


class MemmertOvenNotFoundException(TemperatureControllerNotFoundException):
    def __init__(self, server_url: Optional[str] = None):
        super().__init__(
            server_url or "all",
            "No Memmert Oven available at {port}. Check the LAN connection to the device and ensure that the oven has "
            "the following settings: [IP Address: in the range 192.168.0.0 - 192.168.255.255 except 192.168.100.2], "
            "[Subnet Mask: 255.255.0.0], [Remote Control Mode: enabled and at least set to 'Read'].",
        )


class PermissionDenied(Exception):
    """
    Exception that indicates that the current remote control permissions of the device don't allow accessing a certain
    value
    """

    _source: str
    _expected_permissions: str
    _actual_permissions: str

    def __init__(self, source: str, expected_permissions: str, actual_permissions: str) -> None:
        self._source = source
        self._expected_permissions = expected_permissions
        self._actual_permissions = actual_permissions
        logger.error(f"PermissionDenied {source}, {expected_permissions}, {actual_permissions}")
        super().__init__(
            f"Permission Denied while accessing {source}. Required permissions are "
            f"{self.friendly_expected_permissions} but actual permissions are {self.friendly_actual_permissions}"
        )

    @staticmethod
    def __friendly_permissions(permissions: str) -> list:
        """
        Turns the `permission` string that is in the format "rwa" into a human friendly string like "Read, Write, Alarm"
        """
        res = []
        if permissions.read:
            res += ["Read"]
        if permissions.write:
            res += ["Write"]
        if permissions.alarm:
            res += ["Alarm"]
        return res

    @property
    def source(self) -> str:
        return self._source

    @property
    def expected_permissions(self) -> str:
        return self._expected_permissions

    @property
    def friendly_expected_permissions(self) -> list:
        return self.__friendly_permissions(self._expected_permissions)

    @property
    def actual_permissions(self) -> str:
        return self._actual_permissions

    @property
    def friendly_actual_permissions(self) -> list:
        return self.__friendly_permissions(self._actual_permissions)


class ValueDidNotChange(Exception):
    """
    Exception that indicates that a certain value did not change (probably because the desired value was outside of the
    valid range for this parameter)
    """

    __parameter: _Parameters
    __value: Any

    def __init__(self, parameter: _Parameters, value: Any) -> None:
        self.__parameter = parameter
        self.__value = value

        super().__init__(f"Changing the {parameter.value} to {value} did not change its value.")

    @property
    def parameter(self) -> _Parameters:
        return self.__parameter

    @property
    def value(self) -> str:
        return self.__value

class StrEnum(str, Enum):
    pass


class _Parameters(StrEnum):
    TEMP_SET = "TempSet"
    TEMP_SET_RANGE = "TempSet_Range"
    TEMP1_READ = "Temp1Read"
    AL_TEMP_LO = "AlTempLo"
    AL_TEMP_LO_RANGE = "AlTempLo_Range"
    AL_TEMP_HI = "AlTempHi"
    AL_TEMP_HI_RANGE = "AlTempHi_Range"
    FAN_SET = "FanSet"
    FAN_SET_RANGE = "FanSet_Range"
    FAN_READ = "FanRead"
    FLAP_SET = "FlapSet"
    FLAP_SET_RANGE = "FlapSet_Range"
    CUR_OP = "CurOp"
    SW_REV = "SWRev"
    SN = "SN"
    DEV_TYPE = "DevType"
    RC = "RC"


class _Permissions:
    R: _Permissions  # read permission
    RW: _Permissions  # read + write permission
    RWA: _Permissions  # read + write + alarm permission
    NONE: _Permissions  # no permissions

    __read: bool
    __write: bool
    __alarm: bool

    def __init__(self, permissions: str) -> None:
        if len(permissions) < 3:
            self.__read = False
            self.__write = False
            self.__alarm = False
        else:
            self.__read = permissions[0] == "r"
            self.__write = permissions[1] == "w"
            self.__alarm = permissions[2] == "a"

    @property
    def read(self) -> bool:
        return self.__read

    @property
    def write(self) -> bool:
        return self.__write

    @property
    def alarm(self) -> bool:
        return self.__alarm

    def __lt__(self, o: _Permissions) -> bool:
        return (
            (not self.__read and not self.__write and not self.__alarm and o.__read)
            or (self.__read and not self.__write and not self.__alarm and o.__read and o.__write)
            or (self.__read and self.__write and not self.__alarm and o.__read and o.__write and o.__alarm)
        )

    def __str__(self) -> str:
        return ("r" if self.__read else "-") + ("w" if self.__write else "-") + ("a" if self.__alarm else "-")


_Permissions.R = _Permissions("r--")
_Permissions.RW = _Permissions("rw-")
_Permissions.RWA = _Permissions("rwa")
_Permissions.NONE = _Permissions("")

REQUIRED_PERMISSIONS: Dict[_Parameters, _Permissions] = {
    _Parameters.TEMP_SET: _Permissions.RW,
    _Parameters.TEMP_SET_RANGE: _Permissions.NONE,
    _Parameters.TEMP1_READ: _Permissions.R,
    _Parameters.AL_TEMP_LO: _Permissions.RWA,
    _Parameters.AL_TEMP_LO_RANGE: _Permissions.NONE,
    _Parameters.AL_TEMP_HI: _Permissions.RWA,
    _Parameters.AL_TEMP_HI_RANGE: _Permissions.NONE,
    _Parameters.FAN_SET: _Permissions.RW,
    _Parameters.FAN_SET_RANGE: _Permissions.NONE,
    _Parameters.FAN_READ: _Permissions.R,
    _Parameters.FLAP_SET: _Permissions.RW,
    _Parameters.FLAP_SET_RANGE: _Permissions.NONE,
    _Parameters.CUR_OP: _Permissions.R,
    _Parameters.SW_REV: _Permissions.R,
    _Parameters.SN: _Permissions.R,
    _Parameters.DEV_TYPE: _Permissions.R,
    _Parameters.RC: _Permissions.NONE,
}


class _EventLogCodes(IntEnum):
    """
    Subset of all possible codes for an event log entry
    """

    RESTART = 111
    DATE_TIME_CHANGE = 216
    TEMP_SENSOR_DEFECT = 405
    TEMP_LIMIT_UNDERRUN = 407
    TEMP_LIMIT_EXCEEDED = 408
    ALARM_TEMP_LIMIT_EXCEEDED = 409


class MemmertOvenBase(MemmertOvenInterface):
    """
    The Memmert Oven device driver implementation
    """

    _server_url: str

    _stop_event: Event

    _fan_rotation_speed_set_point: int #: helper for checking if setting the fan speed was successful

    _set_point_value_exception: PermissionDenied
    _actual_temperature_exception: PermissionDenied
    _alarm_temperature_exception: PermissionDenied
    _set_point_range_exception: PermissionDenied
    _alarm_temperature_range_exception: PermissionDenied
    _fan_rotation_speed_exception: PermissionDenied
    _flap_position_exception: PermissionDenied

    def __init__(self, server_url: str):
        """
        Parameters
        ----------
        server_url: str
            The URL of the REST server to use (e.g. "http://192.168.100.100")
        """

        super().__init__()

        self._server_url = server_url
        self._stop_event = Event()

        self._fan_rotation_speed_set_point = None

        self._set_point_value_exception = None
        self._actual_temperature_exception = None
        self._alarm_temperature_exception = None
        self._set_point_range_exception = None
        self._alarm_temperature_range_exception = None
        self._fan_rotation_speed_exception = None
        self._flap_position_exception = None

        try:
            serial_number = self.__get({_Parameters.SN: ""})
            logger.info(f"serial number {serial_number[_Parameters.SN]}")
            # fill all members with initial values
            self.__parse_response(self.__get())
        except (PermissionDenied, KeyError, ConnectionError, ConnectTimeout) as err:
            logger.warning(err, exc_info=err)
            raise MemmertOvenNotFoundException(self._server_url)

    def start(self):
        def poll_api():
            while not self._stop_event.wait(1):
                try:
                    self.__parse_response(self.__get())
                except requests.exceptions.ConnectTimeout as err:
                    logger.warning(err, exc_info=err)
                except:
                    if self._stop_event.is_set():
                        break
                    raise

        def poll_log():
            while not self._stop_event.wait(1):
                self.__parse_log(requests.get("/".join([self._server_url, "Controller", "Config", "log.txt"])))

        Thread(target=poll_api, name="MemmertAPIPollingThread").start()
        Thread(target=poll_log, name="MemmertLogPollingThread").start()

    def stop(self):
        self._stop_event.set()

    def __get(self, params: Optional[Dict[str, Any]] = None) -> Any:
        """
        Performs a GET request on the REST API. The values that should be read from or written to the API have to be
        given as key-value pairs in `params`. Values to be read must have an empty string as value, values to be written
        must have the value that should be written as the value. The key is always the name of the parameter to
        read/write.
        If `params` is `None` then all values of the API will be requested

        Parameters
        ----------
        params:
            The query parameters that will be appended to the request

        Returns
        -------
        The received JSON response or `None` if the request was not successful

        Raises
        ------
        PermissionDenied
            if `params` is not `None` and at least one of the parameters requires higher permissions than currently
            allowed
        """

        def fix_response(text: str) -> str:
            """
            Fixes the returned text response to be a valid JSON string that `json.loads` can parse

            Parameters
            ----------
            text: str
                The response text to fix

            Returns
            -------
            str:
                The fixed response string in valid JSON format
            """
            return f"{{{text[:-2] if text.endswith(', ') else text}}}"

        if self._stop_event.is_set():
            return
        if params is None:
            url = "/".join([self._server_url, "commands.cgi"])
        else:
            self.__check_required_permissions(params.keys())
            url = "/".join([self._server_url, "atmoweb"])
        resp = requests.get(url, params=params, timeout=2)
        logger.debug(f"GET response {resp} {resp.text if resp.ok else None!r})")
        try:
            return json.loads(fix_response(resp.text)) if resp.ok else None
        except json.JSONDecodeError as err:
            logger.error(err, exc_info=err)
            return None

    def __check_required_permissions(self, params: List[_Parameters]) -> None:
        """
        Checks if the current remote control permissions of the device are equal to or higher than the required
        permissions for accessing the given parameters `params`. If there is a parameter in `params` that requires
        higher permissions than what it currently allowed a `PermissionDenied` exception will be raised

        Parameters
        ----------
        params: List[_Parameters]
            The parameters to check the permissions for

        Raises
        ------
        PermissionDenied
            if there is a parameter that requires higher permissions than currently allowed
        """

        if len(params) == 1 and _Parameters.RC in params:
            # we always have permission to get the remote control permissions
            return

        permissions = self._permissions
        for param in params:
            if permissions < REQUIRED_PERMISSIONS[param]:
                raise self.__make_permission_denied(param)

    def __make_permission_denied(self, param: _Parameters) -> PermissionDenied:
        """
        Constructs a `PermissionDenied` exception for the given parameter `param`
        """
        return PermissionDenied(param, REQUIRED_PERMISSIONS[param], self._permissions)

    def __parse_response(self, response: Any) -> None:
        """
        Parses the given JSON response `response` and updates the member variables accordingly

        Parameters
        ----------
        response:
            The JSON response received from the API
        """

        def get_val_or_raise(param: _Parameters) -> Any:
            """
            Returns the value for the given parameter `param` in `response` or raises a `PermissionDenied` if the value
            for `param` indicates that we don't have the required permissions
            """
            resp = response[param]
            if resp == "PermissionDenied":
                raise self.__make_permission_denied(param)
            return resp

        if _Parameters.TEMP_SET in response:
            try:
                self._set_point_value = float(get_val_or_raise(_Parameters.TEMP_SET))
                self._set_point_value_exception = None
            except PermissionDenied as err:
                self._set_point_value_exception = err
        if _Parameters.TEMP1_READ in response:
            try:
                self._actual_temperature = float(get_val_or_raise(_Parameters.TEMP1_READ))
                self._actual_temperature_exception = None
            except PermissionDenied as err:
                self._actual_temperature_exception = err
        if _Parameters.AL_TEMP_LO in response:
            try:
                self._alarm_temperature.min = float(get_val_or_raise(_Parameters.AL_TEMP_LO))
                self._alarm_temperature_exception = None
            except PermissionDenied as err:
                self._alarm_temperature_exception = err
        if _Parameters.AL_TEMP_HI in response:
            try:
                self._alarm_temperature.max = float(get_val_or_raise(_Parameters.AL_TEMP_HI))
                self._alarm_temperature_exception = None
            except PermissionDenied as err:
                self._alarm_temperature_exception = err
        if _Parameters.TEMP_SET_RANGE in response:
            try:
                resp = get_val_or_raise(_Parameters.TEMP_SET_RANGE)
                self._set_point_range.min = float(resp["min"])
                self._set_point_range.max = float(resp["max"])
                self._set_point_range_exception = None
            except PermissionDenied as err:
                self._set_point_range_exception = err
        # AL_TEMP_LO_RANGE and AL_TEMP_HI_RANGE are the same (at least for our oven)
        #   -> if they're ever different we need to change the Feature as well
        if _Parameters.AL_TEMP_LO_RANGE in response:
            try:
                resp = get_val_or_raise(_Parameters.AL_TEMP_LO_RANGE)
                self._alarm_temperature_range.min = float(resp["min"])
                self._alarm_temperature_range.max = float(resp["max"])
                self._alarm_temperature_range_exception = None
            except PermissionDenied as err:
                self._alarm_temperature_range_exception = err
        if _Parameters.AL_TEMP_HI_RANGE in response:
            try:
                resp = get_val_or_raise(_Parameters.AL_TEMP_HI_RANGE)
                self._alarm_temperature_range.min = float(resp["min"])
                self._alarm_temperature_range.max = float(resp["max"])
                self._alarm_temperature_range_exception = None
            except PermissionDenied as err:
                self._alarm_temperature_range_exception = err
        if _Parameters.FAN_SET in response:
            try:
                self._fan_rotation_speed_set_point = int(get_val_or_raise(_Parameters.FAN_SET))
            except PermissionDenied:
                self._fan_rotation_speed_set_point = None
        if _Parameters.FAN_READ in response:
            try:
                self._fan_rotation_speed = int(get_val_or_raise(_Parameters.FAN_READ))
                self._fan_rotation_speed_exception = None
            except PermissionDenied as err:
                self._fan_rotation_speed_exception = err
        if _Parameters.FLAP_SET in response:
            try:
                self._flap_position = int(get_val_or_raise(_Parameters.FLAP_SET))
                self._flap_position_exception = None
            except PermissionDenied as err:
                self._flap_position_exception = err

    def __parse_log(self, response: requests.Response) -> None:
        """
        Parses the given `response` for the log output of the device and updates the `_event_log` member with new
        entries
        """
        if not response.ok:
            logger.warning(f"Failed to parse response {response} because the status is not OK")
            return

        text = response.text

        entries = []
        # first line is header, seconds line is blank
        for line in text.splitlines()[2:]:
            timestamp, entry_type, code, info, checksum = line.split("\t")
            code = int(code)
            if code in (_EventLogCodes.RESTART, _EventLogCodes.DATE_TIME_CHANGE):
                continue
            # timestamp ends with a 'Z' which cannot be parsed by datetime's ISO parser
            timestamp = datetime.datetime.fromisoformat(timestamp[:-1])
            entry_type = EventLogEntryType(entry_type)
            if code == _EventLogCodes.TEMP_LIMIT_UNDERRUN:
                self._min_temperature_underrun = entry_type == EventLogEntryType.EVENT_OCCURRED
                continue
            if code == _EventLogCodes.TEMP_LIMIT_EXCEEDED:
                self._max_temperature_exceeded = entry_type == EventLogEntryType.EVENT_OCCURRED
                continue
            entries += [EventLogEntry(timestamp, entry_type, code, info)]

        self._event_log = entries

    @property
    def _permissions(self) -> _Permissions:
        """
        Returns the current remote control permissions
        """
        return _Permissions(self.__get({_Parameters.RC: ""}).get(_Parameters.RC, ""))

    @property
    def set_point_value(self) -> float:
        """
        Returns the current set point value of the temperature controller
        """
        if self._set_point_value_exception is not None:
            raise self._set_point_value_exception
        return self._set_point_value

    @MemmertOvenInterface.set_point_value.setter
    def set_point_value(self, set_point: float) -> None:
        """
        Set the set point value of the temperature controller
        """
        self.__parse_response(self.__get({_Parameters.TEMP_SET: set_point}))
        if not math.isclose(self._set_point_value, set_point):
            raise ValueDidNotChange(_Parameters.TEMP_SET, set_point)

    @property
    def actual_temperature(self) -> float:
        """
        Returns the current actual temperature of the temperature controller
        """
        if self._actual_temperature_exception is not None:
            raise self._actual_temperature_exception
        return self._actual_temperature

    @property
    def alarm_temperature(self) -> Range:
        """
        Returns the current alarm temperature
        """
        if self._alarm_temperature_exception is not None:
            raise self._alarm_temperature_exception
        return self._alarm_temperature

    @MemmertOvenInterface.alarm_temperature.setter
    def alarm_temperature(self, alarm_temperature: Range) -> None:
        """
        Set the alarm temperatures (minimum and maximum)
        """
        logger.debug(f"set alarm temp {alarm_temperature}")
        self.__parse_response(
            self.__get({_Parameters.AL_TEMP_LO: alarm_temperature.min, _Parameters.AL_TEMP_HI: alarm_temperature.max})
        )

    def set_alarm_temperature_min(self, alarm_temperature_min: float) -> None:
        """
        Set the minimum alarm temperature
        """
        self.alarm_temperature = Range(alarm_temperature_min, self._alarm_temperature.max)
        if not math.isclose(self._alarm_temperature.min, alarm_temperature_min):
            raise ValueDidNotChange(_Parameters.AL_TEMP_LO, alarm_temperature_min)

    def set_alarm_temperature_max(self, alarm_temperature_max: float) -> None:
        """
        Set the maximum alarm temperature
        """
        self.alarm_temperature = Range(self._alarm_temperature.min, alarm_temperature_max)
        if not math.isclose(self._alarm_temperature.max, alarm_temperature_max):
            raise ValueDidNotChange(_Parameters.AL_TEMP_HI, alarm_temperature_max)

    @property
    def set_point_range(self) -> Range:
        """
        Returns the current set point range
        """
        if self._set_point_range_exception is not None:
            raise self._set_point_range_exception
        return self._set_point_range

    @property
    def alarm_temperature_range(self) -> Range:
        """
        Returns the current alarm temperature range
        """
        if self._alarm_temperature_range_exception is not None:
            raise self._alarm_temperature_range_exception
        return self._alarm_temperature_range

    @property
    def fan_rotation_speed(self) -> float:
        """
        Returns the current fan rotation speed
        """
        if self._fan_rotation_speed_exception is not None:
            raise self._fan_rotation_speed_exception
        return self._fan_rotation_speed

    @MemmertOvenInterface.fan_rotation_speed.setter
    def fan_rotation_speed(self, fan_rotation_speed: float) -> None:
        """
        Set the fan rotation speed
        """
        self.__parse_response(self.__get({_Parameters.FAN_SET: fan_rotation_speed}))
        if not math.isclose(self._fan_rotation_speed_set_point, fan_rotation_speed):
            raise ValueDidNotChange(_Parameters.FAN_SET, fan_rotation_speed)

    @property
    def flap_position(self) -> int:
        """
        Returns the current flap position
        """
        if self._flap_position_exception is not None:
            raise self._flap_position_exception
        return self._flap_position

    @MemmertOvenInterface.flap_position.setter
    def flap_position(self, flap_position: float) -> None:
        """
        Set the flap position
        """
        self.__parse_response(self.__get({_Parameters.FLAP_SET: flap_position}))
        if not math.isclose(self._flap_position, flap_position):
            raise ValueDidNotChange(_Parameters.FLAP_SET, flap_position)


class MemmertOven(MemmertOvenBase):
    """
    The Memmert Oven device driver implementation
    """


class MemmertOvenSim(MemmertOvenBase):
    """
    Simulated Memmert Oven device driver
    """

    __port: int = 5000
    __sim_server_app_thread: threading.Thread

    def __init__(self, server_url: str):
        from ..sim.memmert_oven import app

        logging.getLogger("werkzeug").setLevel(logging.WARNING)

        self.__sim_server_app_thread = threading.Thread(
            target=app.run,
            kwargs={"port": self.__port, "use_reloader": False, "threaded": False, "processes": 0},
            daemon=True,
        )
        self.__sim_server_app_thread.start()
        time.sleep(0.5)

        server_url_parts = server_url.split("/")
        server_url_parts[2] = f"localhost:{self.__port}"
        self.__port += 1

        super().__init__("/".join(server_url_parts))


if __name__ == "__main__":
    print("read:", _Permissions.R)
    print("read+write:", _Permissions.RW)
    print("read+write+alarm:", _Permissions.RWA)
    print("none:", _Permissions.NONE)

    assert _Permissions.R < _Permissions.RW
    assert _Permissions.RW < _Permissions.RWA
    assert _Permissions.R < _Permissions.RWA
    assert _Permissions.NONE < _Permissions.R
    assert _Permissions.NONE < _Permissions.RW
    assert _Permissions.NONE < _Permissions.RWA
