from __future__ import annotations

import datetime
from ctypes import Union
from enum import Enum
from typing import Any, Callable, Dict, NamedTuple

from flask import Flask, jsonify, request
from werkzeug.exceptions import NotFound

app = Flask(__name__)


@app.route("/atmoweb", methods=["GET"])
def get_atmoweb():
    return get(request.args)


@app.route("/commands.cgi", methods=["GET"])
def get_all():
    return get({param: "" for param in _Parameters})


class StrEnum(str, Enum):
    pass


class _Parameters(StrEnum):
    TEMP_SET = "TempSet"
    TEMP_SET_RANGE = "TempSet_Range"
    TEMP1_READ = "Temp1Read"
    AL_TEMP_LO = "AlTempLo"
    AL_TEMP_LO_RANGE = "AlTempLo_Range"
    AL_TEMP_HI = "AlTempHi"
    AL_TEMP_HI_RANGE = "AlTempHi_Range"
    FAN_SET = "FanSet"
    FAN_SET_RANGE = "FanSet_Range"
    FAN_READ = "FanRead"
    FLAP_SET = "FlapSet"
    FLAP_SET_RANGE = "FlapSet_Range"
    CUR_OP = "CurOp"
    INFO_TEMP = "InfoTemp"
    INFO_HUM = "InfoHum"
    INFO_VAC = "InfoVac"
    INFO_MSG = "InfoMsg"
    INFO = "Info"
    PROG_CURRENT = "ProgCurrent"
    PROG_DURATION = "ProgDuration"
    PROG_LIST = "ProgList"
    PROG_REMAIN = "ProgRemain"
    TIME = "Time"
    TIME_ZONE = "TimeZone"
    TIME_DST = "TimeDST"
    SW_REV = "SWRev"
    SN = "SN"
    DEV_TYPE = "DevType"
    RC = "RC"


class Range(NamedTuple):
    min: float
    max: float


class CurOp(StrEnum):
    PROGRAM = "Program"
    IDLE = "Idle"
    TIMER = "Timer"
    MANUAL = "Manual"


SEP = ", "


class MemmertOvenSim:
    """
    Simulated Memmert Oven
    """

    temp_set: float = 22
    temp_set_range: Range = Range(20, 300)
    temp1_read: float = 22
    al_temp_lo: float = 10
    al_temp_lo_range: Range = Range(10, 310)
    al_temp_hi: float = 310
    al_temp_hi_range: Range = Range(10, 310)
    fan_set: int = 100
    fan_set_range: Range = Range(0, 100)
    fan_read: int = 100
    flap_set: int = 0
    flap_set_range: Range = Range(0, 100)
    cur_op: CurOp = CurOp.MANUAL
    info_temp: str = None
    info_hum: str = None
    info_vac: str = None
    info_msg: str = None
    info: str = None
    prog_current: str = None
    prog_duration: float = "N/D"
    prog_list: list = []
    prog_remain: float = "N/D"
    time: str
    time_zone: float
    time_dst: int = 1
    sw_rev: str = "04.05.00"
    sn: str = "B266.4908"
    dev_type: str = "UF260"
    rc: str = "rwa"

    @property
    def time(self) -> str:
        return datetime.datetime.now(datetime.timezone.utc).astimezone().strftime("%Y-%m-%dT%H:%M:%S%z")

    @property
    def time_zone(self) -> str:
        return self.time[-5:]

    @staticmethod
    def range_str(range: Range) -> str:
        """
        Serializes a `Range` in a JSON-like string
        """
        return f'{{"min": {range.min:.0f}, "max": {range.max:.0f}}}'

    @staticmethod
    def null_or_str(string: Union[str, None]) -> str:
        """
        Serializes the given `string` in a JSON-like string using 'null' if `string` is `None`
        """
        return "null" if string is None else f'"{string}"'

    @property
    def temp_set_str(self) -> str:
        return f'"{_Parameters.TEMP_SET}": {self.temp_set:.1f}' + SEP

    @property
    def temp_set_range_str(self) -> str:
        return f'"{_Parameters.TEMP_SET_RANGE}": {self.range_str(self.temp_set_range)}' + SEP

    @property
    def temp1_read_str(self) -> str:
        return f'"{_Parameters.TEMP1_READ}": {self.temp1_read:.1f}' + SEP

    @property
    def al_temp_lo_str(self) -> str:
        return f'"{_Parameters.AL_TEMP_LO}": {self.al_temp_lo:.1f}' + SEP

    @property
    def al_temp_lo_range_str(self) -> str:
        return f'"{_Parameters.AL_TEMP_LO_RANGE}": {self.range_str(self.al_temp_lo_range)}' + SEP

    @property
    def al_temp_hi_str(self) -> str:
        return f'"{_Parameters.AL_TEMP_HI}": {self.al_temp_hi:.1f}' + SEP

    @property
    def al_temp_hi_range_str(self) -> str:
        return f'"{_Parameters.AL_TEMP_HI_RANGE}": {self.range_str(self.al_temp_hi_range)}' + SEP

    @property
    def fan_set_str(self) -> str:
        return f'"{_Parameters.FAN_SET}": {self.fan_set:.0f}' + SEP

    @property
    def fan_set_range_str(self) -> str:
        return f'"{_Parameters.FAN_SET_RANGE}": {self.range_str(self.fan_set_range)}' + SEP

    @property
    def fan_read_str(self) -> str:
        return f'"{_Parameters.FAN_READ}": {self.fan_read:.0f}' + SEP

    @property
    def flap_set_str(self) -> str:
        return f'"{_Parameters.FLAP_SET}": {self.flap_set:.0f}' + SEP

    @property
    def flap_set_range_str(self) -> str:
        return f'"{_Parameters.FLAP_SET_RANGE}": {self.range_str(self.flap_set_range)}' + SEP

    @property
    def cur_op_str(self) -> str:
        return f'"{_Parameters.CUR_OP}": {self.null_or_str(self.cur_op)}' + SEP

    @property
    def info_temp_str(self) -> str:
        return f'"{_Parameters.INFO_TEMP}": {self.null_or_str(self.info_temp)}' + SEP

    @property
    def info_hum_str(self) -> str:
        return f'"{_Parameters.INFO_HUM}": {self.null_or_str(self.info_hum)}' + SEP

    @property
    def info_vac_str(self) -> str:
        return f'"{_Parameters.INFO_VAC}": {self.null_or_str(self.info_vac)}' + SEP

    @property
    def info_msg_str(self) -> str:
        return f'"{_Parameters.INFO_MSG}": {self.null_or_str(self.info_msg)}' + SEP

    @property
    def info_str(self) -> str:
        return f'"{_Parameters.INFO}": {self.null_or_str(self.info)}' + SEP

    @property
    def prog_current_str(self) -> str:
        return f'"{_Parameters.PROG_CURRENT}": {self.null_or_str(self.prog_current)}' + SEP

    @property
    def prog_duration_str(self) -> str:
        return f'"{_Parameters.PROG_DURATION}": {self.null_or_str(self.prog_duration)}' + SEP

    @property
    def prog_list_str(self) -> str:
        return f'"{_Parameters.PROG_LIST}": {self.prog_list}' + SEP

    @property
    def prog_remain_str(self) -> str:
        return f'"{_Parameters.PROG_REMAIN}": {self.null_or_str(self.prog_remain)}' + SEP

    @property
    def time_str(self) -> str:
        return f'"{_Parameters.TIME}": {self.null_or_str(self.time)}' + SEP

    @property
    def time_zone_str(self) -> str:
        return f'"{_Parameters.TIME_ZONE}": {self.null_or_str(self.time_zone)}' + SEP

    @property
    def time_dst_str(self) -> str:
        return f'"{_Parameters.TIME_DST}": {self.time_dst:.0f}' + SEP

    @property
    def sw_rev_str(self) -> str:
        return f'"{_Parameters.SW_REV}": {self.null_or_str(self.sw_rev)}' + SEP

    @property
    def sn_str(self) -> str:
        return f'"{_Parameters.SN}": {self.null_or_str(self.sn)}' + SEP

    @property
    def dev_type_str(self) -> str:
        return f'"{_Parameters.DEV_TYPE}": {self.null_or_str(self.dev_type)}' + SEP

    @property
    def rc_str(self) -> str:
        return f'"{_Parameters.RC}": {self.null_or_str(self.rc)}' + SEP


oven = MemmertOvenSim()


def get(params: Dict) -> str:
    def set_val_safe(val: Any, prop: str, conv: Callable) -> None:
        if val:
            try:
                setattr(oven, prop, conv(val))
            except ValueError:
                pass

    resp = ""

    for param, val in params.items():
        if param == _Parameters.TEMP_SET:
            set_val_safe(val, "temp_set", float)
            resp += oven.temp_set_str
        if param in (_Parameters.TEMP_SET, _Parameters.TEMP_SET_RANGE) and _Parameters.TEMP_SET_RANGE not in resp:
            resp += oven.temp_set_range_str
        if param == _Parameters.TEMP1_READ:
            resp += oven.temp1_read_str
        if param == _Parameters.AL_TEMP_LO:
            set_val_safe(val, "al_temp_lo", float)
            resp += oven.al_temp_lo_str
        if param in (_Parameters.AL_TEMP_LO, _Parameters.AL_TEMP_LO_RANGE) and not _Parameters.AL_TEMP_LO_RANGE in resp:
            resp += oven.al_temp_lo_range_str
        if param == _Parameters.AL_TEMP_HI:
            set_val_safe(val, "al_temp_hi", float)
            resp += oven.al_temp_hi_str
        if param in (_Parameters.AL_TEMP_HI, _Parameters.AL_TEMP_HI_RANGE) and not _Parameters.AL_TEMP_HI_RANGE in resp:
            resp += oven.al_temp_hi_range_str
        if param == _Parameters.FAN_SET:
            set_val_safe(val, "fan_set", int)
            resp += oven.fan_set_str
        if param in (_Parameters.FAN_SET, _Parameters.FAN_SET_RANGE) and not _Parameters.FAN_SET_RANGE in resp:
            resp += oven.fan_set_range_str
        if param == _Parameters.FAN_READ:
            resp += oven.fan_read_str
        if param == _Parameters.FLAP_SET:
            set_val_safe(val, "flap_set", int)
            resp += oven.flap_set_str
        if param in (_Parameters.FLAP_SET, _Parameters.FLAP_SET_RANGE) and _Parameters.FLAP_SET_RANGE not in resp:
            resp += oven.flap_set_range_str
        if param == _Parameters.CUR_OP:
            resp += oven.cur_op_str
        if param == _Parameters.INFO_TEMP:
            resp += oven.info_temp_str
        if param == _Parameters.INFO_HUM:
            resp += oven.info_hum_str
        if param == _Parameters.INFO_VAC:
            resp += oven.info_vac_str
        if param == _Parameters.INFO_MSG:
            resp += oven.info_msg_str
        if param == _Parameters.INFO:
            resp += oven.info_str
        if param == _Parameters.PROG_CURRENT:
            resp += oven.prog_current_str
        if param == _Parameters.PROG_DURATION:
            resp += oven.prog_duration_str
        if param == _Parameters.PROG_LIST:
            resp += oven.prog_list_str
        if param == _Parameters.PROG_REMAIN:
            resp += oven.prog_remain_str
        if param == _Parameters.TIME:
            resp += oven.time_str
        if param == _Parameters.TIME_ZONE:
            resp += oven.time_zone_str
        if param == _Parameters.TIME_DST:
            resp += oven.time_dst_str
        if param == _Parameters.SW_REV:
            resp += oven.sw_rev_str
        if param == _Parameters.SN:
            resp += oven.sn_str
        if param == _Parameters.DEV_TYPE:
            resp += oven.dev_type_str
        if param == _Parameters.RC:
            resp += oven.rc_str
        if param not in list(_Parameters):
            resp += f"{param}=unknown"
    return resp


if __name__ == "__main__":
    app.run(debug=True)
