"""
A device driver implementation for the serial interface of 2mag MIXdrive stirrers

Author
------
Florian Meinicke (florian.meinicke@cetoni.de)

Date
----
21.07.2022
"""

from __future__ import annotations

import logging
import re
import time
from enum import Enum
from queue import Empty, Queue
from threading import Event, Thread
from typing import Optional

from .abc import (
    SerialStirringDeviceInterface,
    StirringDeviceInterface,
    StirringDeviceNotFoundException,
    _SerialStirringDeviceReader,
)

TERMINATOR = "\r"  # terminator as per manual
_SerialStirringDeviceReader.TERMINATOR = TERMINATOR.encode("utf-8")

logger = logging.getLogger(__name__)


class MIXdriveNotFoundException(StirringDeviceNotFoundException):
    def __init__(self, port: Optional[str] = None):
        super().__init__(port, "No 2mag MIXdrive detected on {port}")


class MIXdriveCommand:
    """
    A MIXdrive command
    """

    # the commands that don't require a payload for convenience
    START: MIXdriveCommand
    STOP: MIXdriveCommand
    SEND_RPM: MIXdriveCommand
    SEND_POWER: MIXdriveCommand
    SET_DEFAULT: MIXdriveCommand
    SEND_STATUS: MIXdriveCommand

    class Command(Enum):
        """
        All the commands that the 2mag MIXdrive understands
        """

        START = "start"
        STOP = "stop"
        SET_RPM = "setrpm"
        SEND_RPM = "sendrpm"
        SET_POWER = "setpower"
        SEND_POWER = "sendpower"
        SET_ADD = "setadd"
        SET_DEFAULT = "setdefault"
        SEND_STATUS = "sendstatus"

    __command: Command
    __payload: Optional[str]
    __address: Optional[str]

    def __init__(self, command: Command, payload: Optional[str] = None, address: Optional[str] = None) -> None:
        self.__command = command
        if payload is not None and not isinstance(payload, str):
            raise TypeError("Payload of MIXdrive command can only be str or None!")
        self.__payload = payload
        if address is not None and len(address) > 2:
            raise ValueError(
                "Address for MIXdrive commands can only be 1 (addresses single device) or 2 (addresses device range) "
                "characters!"
            )
        self.__address = address

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__command}, {self.__payload}, {self.__address})"

    def __str__(self) -> str:
        s = f"{self.__command.value}"
        if self.__payload is not None:
            s += f"_{self.__payload}"
        if self.__address is not None:
            s += f"_{self.__address}"
        return s


MIXdriveCommand.START = MIXdriveCommand(MIXdriveCommand.Command.START)
MIXdriveCommand.STOP = MIXdriveCommand(MIXdriveCommand.Command.STOP)
MIXdriveCommand.SEND_RPM = MIXdriveCommand(MIXdriveCommand.Command.SEND_RPM)
MIXdriveCommand.SEND_POWER = MIXdriveCommand(MIXdriveCommand.Command.SEND_POWER)
MIXdriveCommand.SET_DEFAULT = MIXdriveCommand(MIXdriveCommand.Command.SET_DEFAULT)
MIXdriveCommand.SEND_STATUS = MIXdriveCommand(MIXdriveCommand.Command.SEND_STATUS)


class MIXdriveError(Exception):
    """
    An error from the MIXdrive controller
    """

    __err_code: str
    __msg: str

    def __init__(
        self,
        err_code: str,
        *,
        command: Optional[str] = None,
        operation: Optional[str] = None,
        param: Optional[str] = None,
        extra_msg: Optional[str] = None,
    ) -> None:
        """
        Create a MIXdriveError

        Parameters
        ----------
        err_code: str
            The error code identifying the error type

        Keyword Arguments
        -----------------
        command: str
            The command that is unknown in case this is an "Unknown Command" error
        operation: str
            The operation that is currently not possible in case this is a "Manual mode" error
        param: str
            The parameter of that is out of range in case this is a "Parameter out of range" error
        """
        self.__err_code = err_code
        if self.is_unknown_command_error:
            self.__msg = "Unknown command" + f": {command}" if command is not None else ""
        elif self.is_manual_mode_error:
            self.__msg = "Manual mode" + f": {operation} not possible" if operation is not None else ""
        elif self.is_out_of_range_error:
            self.__msg = "Parameter out of range" + f": {param} is out of range" if param is not None else ""
        self.__extra_msg = " " + extra_msg if extra_msg is not None else ""
        super().__init__(self.__msg + self.__extra_msg)

    @property
    def err_code(self) -> str:
        return self.__err_code

    @property
    def msg(self) -> str:
        return self.__msg

    @property
    def extra_msg(self) -> str:
        return self.__extra_msg

    @property
    def is_unknown_command_error(self) -> str:
        return self.__err_code == "1"

    @property
    def is_manual_mode_error(self) -> str:
        return self.__err_code == "2"

    @property
    def is_out_of_range_error(self) -> str:
        return self.__err_code == "3"


class MIXdrive(SerialStirringDeviceInterface):
    """
    Class for serial communication to a 2mag MIXdrive
    """

    not_found_exception = MIXdriveNotFoundException

    __write_queue: Queue[MIXdriveCommand]
    __error_queue: Queue[MIXdriveError]
    __data_received_event: Event

    # __status_regex = re.compile(r"(?<=(?:OK|ER)_V\d{4}_)(REM|MAN|OFF)(?=_[A-Z]{1})")
    # (OK|ER)_(?:(?P<version>V\d{4})_)?(?(version)(REM|OFF|MAN)|(\w+))(?=_[A-Z]{1})
    __status_regex = re.compile(
        r"(OK|ER)_(?:(?P<version>V\d{4})_)?(?(version)(REM|OFF|MAN)|(?<=ER_)([A-Z0-9]+))(?=_[A-Z]{1})"
    )
    # (?<=(?:OK|ER)_)(\d{3,4})(?=RPM_[A-Z]{1})
    __rpm_regex = re.compile(r"(OK|ER)_(?:_(\d{1})_)?(\d{3,4})(?=RPM_[A-Z]{1})")
    # (?<=(?:OK|ER)_POWER)(\d{3,4})(?=_[A-Z]{1})
    __power_regex = re.compile(r"(OK|ER)_(?:_(\d{1})_)?(?:POWER)(\d{3,4})(?=_[A-Z]{1})")

    def __init__(self, port: Optional[str] = None):
        super().__init__(port)

        self.__write_queue = Queue()
        self.__error_queue = Queue()
        self.__data_received_event = Event()

    def start(self):
        super().start()

        def poll():
            while self._protocol.transport.alive:
                try:
                    self.__write_command_and_wait(self.__write_queue.get(block=True, timeout=1))
                    # this will notify observers that are waiting for the queue to become empty with `Queue.join()`
                    self.__write_queue.task_done()
                except Empty:
                    self.__write_queue.put(MIXdriveCommand.SEND_STATUS)
                    self.__write_queue.put(MIXdriveCommand.SEND_RPM)
                    self.__write_queue.put(MIXdriveCommand.SEND_POWER)

        # fill all members with default values
        self.__write_command_and_wait(MIXdriveCommand.SEND_STATUS)
        self.__write_command_and_wait(MIXdriveCommand.SEND_RPM)
        self.__write_command_and_wait(MIXdriveCommand.SEND_POWER)
        # start polling for value changes
        Thread(target=poll, name="MIXdrivePollingThread").start()

    def __write_command_and_wait(self, command: MIXdriveCommand) -> None:
        """
        Writes the given `command` to the device and waits until the response has been received

        If the response has not been received after 2 seconds it means that either the response has been lost or the
        device dis not send a response. Usually, the response should be received after 100ms (as per device manual).
        """
        logger.debug(f"Writing {command!r}")
        self._protocol.write_line(str(command))
        self.__data_received_event.wait(2)
        self.__data_received_event.clear()

    def __send_command(self, command: MIXdriveCommand) -> None:
        """
        Sends the given `command` to the device
        """
        logger.debug("waiting for queue empty")
        self.__write_queue.join()  # wait for all commands in the queue to be processed
        logger.debug(f"queue empty -> adding command {command}")
        self.__write_queue.put(command)

    @StirringDeviceInterface.rpm.setter
    def rpm(self, rpm: float) -> None:
        if not rpm.is_integer():
            raise MIXdriveError(err_code="3", param="RPM", extra_msg="(the value has to be a positive integer)")
        self.__send_command(MIXdriveCommand(MIXdriveCommand.Command.SET_RPM, f"{int(rpm):04}"))
        self.__data_received_event.wait(2)
        try:
            raise self.__error_queue.get_nowait()
        except Empty:
            logger.debug(f"Setting rpm to {rpm} (no error in queue)")
            self._rpm = rpm

    @StirringDeviceInterface.power.setter
    def power(self, power: float) -> None:
        if not power.is_integer():
            raise MIXdriveError(
                err_code="3",
                param="power",
                extra_msg=(
                    "(the value has to be a positive integer, either 25/50/75/100 if you're using MIXcontrol 20 or "
                    "10...100 with a step of 10 if your're using MIXcontrol 40)"
                ),
            )
        self.__send_command(MIXdriveCommand(MIXdriveCommand.Command.SET_POWER, f"{int(power):02}"))
        self.__data_received_event.wait(2)
        try:
            raise self.__error_queue.get_nowait()
        except Empty:
            logger.debug(f"Setting power to {power} (no error in queue)")
            self._power = power

    def start_stirring(self) -> None:
        self.__send_command(MIXdriveCommand.START)
        self.__data_received_event.wait(2)
        try:
            raise self.__error_queue.get_nowait()
        except Empty:
            pass

    def stop_stirring(self) -> None:
        self.__send_command(MIXdriveCommand.STOP)
        self.__data_received_event.wait(2)
        try:
            raise self.__error_queue.get_nowait()
        except Empty:
            pass

    def on_serial_data_received(self, data: str) -> None:
        """
        Function to handle the `data` read from the stirring device
        """
        logger.debug(data)

        def report_possible_error(ok: str, err_code: str, **kwargs) -> bool:
            """
            Reports a possible error (indicated by `ok == "ER"`) by putting the corresponding `MIXdriveError` into the
            `__error_queue` and returns whether `ok` indicated an error

            Parameters
            ----------
            ok: str
                Indicates if there is an error, either "OK" or "ER"
            err_code: str
                The error code to construct the `MIXdriveError` from
            **kwargs: dict
                Additional arguments to pass on to `MIXdriveError`
            """
            if ok == "ER":
                logger.debug(f"Reporting error with code {err_code} and args {kwargs}")
                self.__error_queue.put(MIXdriveError(err_code, **kwargs))
                return True
            return False

        # OK_V8360_OFF_A / OK_V8360_REM_A / OK_V8360_MAN_A -> mode OFF / REM / MAN, SW version V8360
        # ER_UNKNOWNCOMMAND_A -> unknown command
        status_match = self.__status_regex.findall(data)
        # OK_0350RPM_A -> ok, RPM is 350
        # ER__3_0500RPM_A -> error, RPM out of range, it's still 500
        rpm_match = self.__rpm_regex.findall(data)
        # OK_POWER050_A -> ok, power is 50
        # ER__3_POWER050_A -> error, power out of range, it's still 50
        power_match = self.__power_regex.findall(data)
        logger.debug(f"status {status_match}, rpm {rpm_match}, power {power_match}")
        if status_match:
            ok, _, mode, rest = status_match[0]
            if not report_possible_error(ok, "1" if rest == "UNKNOWNCOMMAND" else rest):
                self._is_stirring = mode == "REM"
            logger.debug(f"is_stirring {self._is_stirring}")
        if rpm_match:
            ok, err_code, rpm = rpm_match[0]
            if (
                not report_possible_error(
                    ok, err_code, param="RPM", extra_msg="(the value has to be a positive integer)"
                )
                and self._is_stirring
            ):
                # only if the mode is "REM" the device reports the correct value, if the mode is "OFF" it reports the
                # value for the manual mode
                logger.debug(f"Setting rpm to {rpm} (ok: {ok}, is_stirring: {self._is_stirring})")
                self._rpm = float(rpm)
            logger.debug(f"rpm {self._rpm}")
        if power_match:
            ok, err_code, power = power_match[0]
            if (
                not report_possible_error(
                    ok,
                    err_code,
                    param="power",
                    extra_msg=(
                        "(the value has to be a positive integer, either 25/50/75/100 if you're using MIXcontrol 20 or "
                        "10...100 with a step of 10 if your're using MIXcontrol 40)"
                    ),
                )
                and self._is_stirring
            ):
                # only if the mode is "REM" the device reports the correct value, if the mode is "OFF" it reports the
                # value for the manual mode
                logger.debug(f"Setting power to {power} (ok: {ok}, is_stirring: {self._is_stirring})")
                self._power = float(power)
            logger.debug(f"power {self._power}")

        self.__data_received_event.set()

    @property
    def unique_stirring_device_identifier_request(self) -> str:
        """
        This function should return a string containing a request for the stirring device to return some kind of
        unique identifier or something that can be used to identify if the current serial connection uses the correct
        stirring device (e.g. this could be a request to return the serial number of the stirring device).
        The response of the device will be passed to the `is_valid_stirring_device()` function to validate it.
        """
        return "sendstatus"

    @staticmethod
    def is_valid_stirring_device(data: bytes) -> bool:
        """
        This function should return whether `data` indicates that the current serial connection is using the correct
        stirring device.

        :param data: The serial data which was read as a response to writing the
                     `unique_stirring_device_identifier_request`
        """
        # b"OK_V8360_OFF_A\r"
        logger.debug(data)
        logger.debug(MIXdrive.__status_regex.findall(data.decode("utf-8")))
        return bool(MIXdrive.__status_regex.findall(data.decode("utf-8")))


class MIXdriveSim(StirringDeviceInterface):
    """
    A simple simulated device driver implementation for 2 mag MIXdrive devices
    """

    def __init__(self, *args, **kwargs):
        super().__init__()

    @StirringDeviceInterface.rpm.setter
    def rpm(self, rpm: float) -> None:
        if not rpm.is_integer():
            raise MIXdriveError("3", param="RPM")
        self._rpm = rpm

    @StirringDeviceInterface.power.setter
    def power(self, power: float) -> None:
        if not power.is_integer():
            raise MIXdriveError("3", param="power")
        self._power = power

    def start_stirring(self) -> None:
        self._is_stirring = True

    def stop_stirring(self) -> None:
        self._is_stirring = False

    def start(self):
        pass

    def stop(self):
        pass


# ----------------------------------------------------------------------------
# test
if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s [%(threadName)-12.12s] %(levelname)-8s| %(name)s %(module)s.%(funcName)s: %(message)s",
        level=logging.DEBUG,
    )

    mixdrive = MIXdrive()
    mixdrive.open()
    print("is_stirring", mixdrive.is_stirring)
    print("rpm", mixdrive.rpm)
    print("power", mixdrive.power)
    mixdrive.rpm = 1000
    mixdrive.power = 50
    mixdrive.start_stirring()
    # time.sleep(2)
    print("is_stirring", mixdrive.is_stirring)
    print("rpm", mixdrive.rpm)
    print("power", mixdrive.power)
    input("Press enter to stop stirring")
    mixdrive.stop_stirring()
    while mixdrive.is_stirring:
        time.sleep(0.1)
    mixdrive.stop()
